#define USE_TI89
#define USE_TI92P
#define USE_V200

#include <fat.h>

//-----------------------------------------------------------------------------
// include the utility routines ...
//-----------------------------------------------------------------------------
#include <fatutil.c>

enum COMPTYPE { NONE=0x00, RAND=0x01, STRAT=0x02, FIGHT=0x04 };

FATCONFIG fc;
BOOL o_sprtavail = TRUE;
BOOL o_night = FALSE;
short o_comp = NONE;
BOOL o_timer = TRUE;

char mazename[9];
unsigned long besttime = 0;
BOOL loaded = FALSE;
//short l_xpos = -1, l_ypos = -1, l_o = -1;

short display_offsetx = 0; // will be set depending on calctype (TI89 or TI92p)
short display_offsety = 0; // will be set depending on calctype (TI89 or TI92p)

//=============================================================================
//=============================================================================
// Utility Functions
//=============================================================================
//=============================================================================

void InputStr (char *buffer, short maxlen)
{
  SCR_STATE ss;
  short key, i = 0;
  buffer[0] = 0;
  SaveScrState (&ss);
  do
    {
      MoveTo (ss.CurX, ss.CurY);
      printf ("%s_  ", buffer);
        // Two spaces are required if F_4x6 font is used
      key = ngetchx ();
//      if (key >= ' ' && key <= '~' && i < maxlen) buffer[i++] = key;
      if (isalnum(key) && i < maxlen) buffer[i++] = key;
      if (key == KEY_BACKSPACE && i) i--;
      buffer[i] = 0;
    } while (key != KEY_ENTER);
}

void Clear()
{
	int i;
	for( i=0; i<2; i++ ){
		SetPlane( i );
		clrscr();
	}
}


//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
// Functions related to sprites
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------

inline unsigned char GetMapValue( int x, int y );

short GetDirs( short dirs[], short x, short y, short o )
{
	short n = 0;
	short i;
	
	if( o )	o -= ANGLE_090;
	else		o = ANGLE_270;
	
	for( i=0; i<4; i++ ){
		if( o == ANGLE_NORTH && !GetMapValue( x, y-1 ) )	dirs[ n++ ] = 0;
		else if( o == ANGLE_EAST	&& !GetMapValue( x+1, y ) ) dirs[ n++ ] = 1;
		else if( o == ANGLE_SOUTH	&& !GetMapValue( x, y+1 ) ) dirs[ n++ ] = 2;
		else if( o == ANGLE_WEST	&& !GetMapValue( x-1, y ) )	dirs[ n++ ] = 3;
		
		o += ANGLE_090;
		if( o == ANGLE_360 ) o = 0;
	}
	return n;
}

void MoveSprite( )
{
	short x = (fc.sprites->xpos),
				y = (fc.sprites->ypos),
				o = fc.sprites->orientation;
	
	if( o_comp == FIGHT ){
		int i;
    short stepx;
    short stepy;
    short newxpos;
    short newypos;
    short newxmod;
    short newymod;
    
    short xdiff, ydiff;

		for( i=0; i<fc.nr_sprites; i++ ){
			xdiff = fc.cam_xpos - fc.sprites[i].xpos;
			ydiff = fc.cam_ypos - fc.sprites[i].ypos;
			
			if( abs(xdiff) < 64 && abs(ydiff) < 64 ) continue;

			fc.sprites[i].orientation = FAT_GetArcTanYX( ydiff, xdiff);

			stepx   = FAT_Cos16384(fc.sprites[i].orientation) >> 11;
			stepy   = FAT_Sin16384(fc.sprites[i].orientation) >> 11;
			newxpos = fc.sprites[i].xpos + stepx;
			newypos = fc.sprites[i].ypos + stepy;
			newxmod = newxpos >> 6;
			newymod = newypos >> 6;

	    //-------------------------------------------------------------------------
	    // check if we are still within the map
	    //-------------------------------------------------------------------------
			if (newxmod > fc.map_width-1  ||
				newymod > fc.map_height-1 ||
				newxmod < 0 || newymod < 0)
			{
				continue;
			}

	    //-------------------------------------------------------------------------
	    // check if we would move into a solid square
	    //-------------------------------------------------------------------------
			if (*(fc.map_data+newymod*fc.map_width+newxmod)>0) continue;

	    //-------------------------------------------------------------------------
	    // everything okay -> let's set the new position
	    //-------------------------------------------------------------------------
			fc.sprites[i].xpos = newxpos;
			fc.sprites[i].ypos = newypos;
		}
	}
	else{
		if( x%64 != 32 || y%64 != 32 ){
			switch( o ){
				case( ANGLE_NORTH ): fc.sprites->ypos-=4; break;
				case( ANGLE_SOUTH ): fc.sprites->ypos+=4; break;
				case( ANGLE_WEST ):  fc.sprites->xpos-=4; break;
				default:						 fc.sprites->xpos+=4;
			}
			return;
		}
		else{
			short ndir;
			short dirs[4]/* = {-1, -1, -1, -1 }*/;
			ndir = GetDirs( dirs, x/64, y/64, o );
	
			if( o_comp == RAND ){
				switch( dirs[ ndir < 3? 0: random( ndir-1 ) ] ){
					case( 0 ): fc.sprites->orientation = ANGLE_NORTH; break;
					case( 1 ): fc.sprites->orientation = ANGLE_EAST; break;
					case( 2 ): fc.sprites->orientation = ANGLE_SOUTH; break;
					default: fc.sprites->orientation = ANGLE_WEST;
				}
				switch( fc.sprites->orientation ){
					case( ANGLE_NORTH ): fc.sprites->ypos-=4; break;
					case( ANGLE_SOUTH ): fc.sprites->ypos+=4; break;
					case( ANGLE_WEST ):  fc.sprites->xpos-=4; break;
					default:	 fc.sprites->xpos+=4;
				}
			}
			else
			{
				switch( dirs[0] ){
					case( 0 ): fc.sprites->orientation = ANGLE_NORTH; break;
					case( 1 ): fc.sprites->orientation = ANGLE_EAST; break;
					case( 2 ): fc.sprites->orientation = ANGLE_SOUTH; break;
					default: fc.sprites->orientation = ANGLE_WEST;
				}
				switch( fc.sprites->orientation ){
					case( ANGLE_NORTH ): fc.sprites->ypos-=4; break;
					case( ANGLE_SOUTH ): fc.sprites->ypos+=4; break;
					case( ANGLE_WEST ):  fc.sprites->xpos-=4; break;
					default:	 fc.sprites->xpos+=4;
				}
			}
		}
	}
}

void Shoot()
{
	
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
// Functions related to the GUI
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------

void DrawGUI()
{	
	static unsigned long empty[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	memset(GetPlane(LIGHT_PLANE),0x0,LCD_SIZE);
	memset(GetPlane(DARK_PLANE),0xff,LCD_SIZE);

	Sprite32(158-32-14+display_offsetx,display_offsety+2+3,16,empty,GetPlane(0),SPRT_AND);
	Sprite32(158-32-14+display_offsetx,display_offsety+2+3,16,FAT_GetTinyLogo(),GetPlane(0),SPRT_OR);
	Sprite32(158-32-14+display_offsetx,display_offsety+2+3,16,empty,GetPlane(1),SPRT_AND);
	Sprite32(158-32-14+display_offsetx,display_offsety+2+3,16,FAT_GetTinyLogo()+16,GetPlane(1),SPRT_OR);
	
	SetPlane( DARK_PLANE );
	FontSetSys( F_8x10 );
	DrawStrXY( 97+14, 25+3, "Maze", A_XOR);
 	DrawStrXY( 97+5, 40+3, "Runner", A_XOR);

	if( o_timer ){
		DrawStrXY( 97+10, 61+3, "Time:", A_XOR);
	}
}

//=============================================================================
// initializes GUI (darkgray background, tiny logo and counter)
//=============================================================================
void InitGUI(COUNTER *counter) {
	DrawGUI();

	InitCounter(counter,/*158-6*5*/97+12+display_offsetx,/*99-10*/76+display_offsety+3,7,/*0*/fc.frame_counter);
	if(o_timer)	DrawCounter(counter);
}


//=============================================================================
// output frame count ever 10th frame
//=============================================================================
void frame_drawn(FATCONFIG* fc,COUNTER* counter) {
	short i;
	if (o_timer && fc->frame_counter%10==0) {
		// we starting incrementing at 2nd position and not on first!
		for (i=counter->length-2;i>=0;i--) {
			counter->newcount[i] = counter->newcount[i]+1;
			if (counter->newcount[i] < 10) break;
			else                           counter->newcount[i]=0;
		}
		DrawCounter(counter);
	}
}

//-------------------------------------------------------------------------
// Output title
//-------------------------------------------------------------------------

void OutputTitle(void) {
	short  i;
   
	for (i=0;i<2;i++) {
		SetPlane(i);
	        
		FontSetSys( F_8x10 );
		DrawStrXY( 97, 20, "Maze", A_REPLACE);
	 	DrawStrXY( 97, 35, "Runner", A_REPLACE);
	
		if( o_timer ){
			FontSetSys( F_4x6 );
			DrawStrXY( 97, 61, "Your Time", A_REPLACE);
		}
	}
}

//-------------------------------------------------------------------------
// Print current time in maze 1 = 10 frames.
//-------------------------------------------------------------------------

void PrintTime( int time )
{
	if( o_timer && time%10 == 0 )
		printf_xy( 97, 67, "%05i", time/10 );
}

//=============================================================================
// our Predraw callback which simulates the fat-in-the-distance effect by
// modifying the textures used to render the view
//=============================================================================
void PredrawCallback(FATCONFIG* fc,DEPTHDATA* dd) {
	short i;

	fc = fc; // just to make the compiler happy. otherwise it will complain
           // about the unused parameter fc

	for (i=0;i<ANGLE_060;i++){
		if (dd->distance[i] < 76) continue;
		if (dd->distance[i] < 100){
			dd->lightdata[i<<1]     = 0xffffffffL;
			dd->lightdata[(i<<1)+1] = 0xffffffffL;
		}
		else if (dd->distance[i] < 170){
			dd->darkdata[i<<1]     = 0xffffffffL;
			dd->darkdata[(i<<1)+1] = 0xffffffffL;
		}
		else{
			dd->lightdata[i<<1]     = 0xffffffffL;
			dd->lightdata[(i<<1)+1] = 0xffffffffL;
			dd->darkdata[i<<1]      = 0xffffffffL;
			dd->darkdata[(i<<1)+1]  = 0xffffffffL;
		}
	}
}

//=============================================================================
//=============================================================================
//  Functions related to the map
//=============================================================================
//=============================================================================

//=============================================================================
// gets the value of a spot on the map give the x and y coordinates
//=============================================================================

inline unsigned char GetMapValue( int x, int y )
{
	return fc.map_data[ x + fc.map_width*y ];
}

//=============================================================================
//  Print MiniMap
//=============================================================================

void PrintMiniMap( int x, int y )
{
	int i, j;
	
	for( i=0; i<fc.map_height; i++ ){
		for( j=0; j<fc.map_width; j++ ){
			if( GetMapValue(j,i) == 1 || GetMapValue(j,i) == 2 ){
				DrawPix( x+j, y+i, A_NORMAL );
			}
		}
	}
}

//-------------------------------------------------------------------------
// Display MiniMap (on key press)
//-------------------------------------------------------------------------
void DisplayMiniMap(int x, int y, COUNTER* frames_display )
{
	unsigned int i;
	BOOL on = 0;
	
	for(i=0; i<2; i++){
		SetPlane( i );
		clrscr();
		if( fc.map_width <= 158 && fc.map_height <= 76 ){
			PrintMiniMap( 79-fc.map_width/2, 48-fc.map_height/2 );
		}
		else{
			FontSetSys( F_8x10 );
			printf_xy( 30,34, "Map Too Large" );
		}
	}
	SetPlane( DARK_PLANE );
	for( i=0; FAT_GetPadState() & PADSTATE_APPS; i++ ){
		if( i == 2000 ){
			i = 0;
			if( on )
				DrawPix((79-fc.map_width/2)+(x/64),
								(48-fc.map_height/2)+(y/64), A_REVERSE);
			else
				DrawPix((79-fc.map_width/2)+(x/64),
								(48-fc.map_height/2)+(y/64), A_NORMAL);
			on = on? 0: 1;
		}
	}
	Clear();

	frames_display->already_used = 0;
	DrawGUI();
	frame_drawn(&fc, frames_display );
}

/* genmaze.c: random maze generation for TI-89 / 92 Plus
 */
 
//#include <tigcclib.h>

//GenMaze: random maze generation
//maze: pre-alloacted char array
//width & height: MUST BE ODD. Also note walls are 1 tile wide.
//empty & wall: chars to use for empty space and wall spaces
void GenMaze(char* maze, int width, int height, char empty, char wall)
{
	int i, j;
	int* open = malloc((width - 1) / 2 * ((height - 1) / 2) * sizeof(int));
	char* openMap = calloc(height, width);
	int numOpen = 1;
	int yMove, xMove;
	int rndOpen;
	/* init to every other square empty */
	for(i = 0; i < height; i++)
		for(j = 0; j < width; j++)
			if((j & 1) && (i & 1)) /* quick odd */
				maze[i * width + j] = empty;
			else
				maze[i * width + j] = wall;
	open[0] = width + 1; /* upper left */
	openMap[width + 1] = 1;
	
	randomize();
	while(numOpen < ((width - 1) / 2) * ((height - 1) / 2)) {
		switch(random(4))
		{
		case 0:
			yMove = -1;
			xMove = 0;
			break;
		case 1:
			yMove = 0;
			xMove = 1;
			break;
		case 2:
			yMove = 1;
			xMove = 0;
			break;
		case 3:
			yMove = 0;
			xMove = -1;
		}
		rndOpen = random(numOpen);
		if(open[rndOpen] % width + (2 * xMove) < 0 ||
		   open[rndOpen] % width + (2 * xMove) >= width ||
		   open[rndOpen] / width + (2 * yMove) < 0 ||
		   open[rndOpen] / width + (2 * yMove) >= height)
			continue;
		if(openMap[open[rndOpen] + (yMove * 2) * width + (xMove * 2)] == 0) {
			openMap[open[rndOpen] + (yMove * 2) * width + (xMove * 2)] = 1;
			maze[open[rndOpen] + yMove * width + xMove] = empty;
			open[numOpen++] = open[rndOpen] + (yMove * 2) * width + (xMove * 2);
		}
	}
	free(open);
	free(openMap);
	
//	fc.map_data[1] = 2;
	fc.map_data[fc.map_width] = 2;
	fc.map_data[fc.map_width*(fc.map_height-1)-1] = 0;
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
// Functions related to files
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Save maze as name.maze
//-------------------------------------------------------------------------

BOOL SaveMap( const char* name )
{
	int i;
	FILE *fp = fopen( name, "wb" );
	if( ! fp ){
		printf( "Maze could not be saved" );
		while( !_rowread(0) );
		return 0;
	}
	short fcBuf[6] = { fc.map_width, fc.map_height, fc.nr_sprites, fc.cam_xpos,
											 fc.cam_ypos, fc.cam_orientation };
	BOOL oBuf[2] = { o_night, o_timer };
	short sprtBuf[3];

	fwrite( fcBuf, sizeof(short), 6, fp );
	fwrite( &fc.frame_counter, sizeof(long), 1, fp );
	fwrite( oBuf, sizeof(BOOL), 2, fp );
	fwrite( &o_comp, sizeof(short), 1, fp);
	for( i=0; i<fc.nr_sprites; i++ ){
		sprtBuf[0] = fc.sprites[i].xpos;
		sprtBuf[1] = fc.sprites[i].ypos;
		sprtBuf[2] = fc.sprites[i].orientation;

		fwrite( sprtBuf, sizeof(short), 3, fp );
	}
	fwrite( fc.map_data, sizeof(char), fc.map_width*fc.map_height, fp );

	fputc( 0, fp );
	fputs( "MAZE", fp );
	fputc( 0, fp );
	fputc( OTH_TAG, fp );
	fclose( fp );

	return 0;
}

//-------------------------------------------------------------------------
// Load maze name
//-------------------------------------------------------------------------

BOOL LoadMap( const char* name, LCD_BUFFER lcd )
{
	FILE *fp = fopen( name, "rb" );
	short fcBuf[6];
	BOOL oBuf[2];
	short sprtBuf[3];
	int i;
	
	if( !fp )
		return 1;
	loaded = TRUE;
	
	fread( fcBuf, sizeof(short), 6, fp );
	fc.map_width				= fcBuf[0];
	fc.map_height				= fcBuf[1];
	fc.nr_sprites				= fcBuf[2];
	fc.cam_xpos					= fcBuf[3];
	fc.cam_ypos					= fcBuf[4];
	fc.cam_orientation	= fcBuf[5];
	
	fread( &fc.frame_counter, sizeof(long), 1, fp );
	
	fread( oBuf, sizeof(BOOL), 2, fp );
	o_night = oBuf[0];
	o_timer = oBuf[1];

	fread( &o_comp, sizeof(short), 1, fp );

	fc.sprites = calloc( fc.nr_sprites, sizeof(FATSPRITE) );
	for( i=0; i<fc.nr_sprites; i++ ){
		fread( sprtBuf, sizeof(short), 3, fp );

		FAT_InitSprite(fc.sprites+i);
		fc.sprites[i].xpos = sprtBuf[0];
		fc.sprites[i].ypos = sprtBuf[1];
		fc.sprites[i].orientation = sprtBuf[2];
	}
	
	fc.map_data = calloc( fc.map_width*fc.map_height, sizeof(char) );
	if( !fc.map_data ){
		return 1;
	}
	fread( fc.map_data, sizeof(char), fc.map_width*fc.map_height, fp );
	
	return 0;
}

//-------------------------------------------------------------------------
// Save maze as:
//-------------------------------------------------------------------------

void Save(COUNTER *frames_display)
{
	char name[9];
	unsigned long time = fc.frame_counter;
	fc.frame_counter = ULONG_MAX;
	
	clrscr();
  FAT_AllKeysReleased();
  GrayMode(GRAY_OFF);
  RestoreHandlerForINTs();
	
	FontSetSys( F_8x10 );
	do{
		printf( "Save map as:\n" );
		InputStr( name, 8 );
		if( isdigit( name[0] ) )
			printf("\nFirst must be a letter\n");
	}while( isdigit(name[0]) );
	if(name[0] != '\0')
		SaveMap( name );
	
	fc.frame_counter = time;
	InstallDummyHandlerForINTs();
	GrayMode( GRAY_ON );
	
	frames_display->already_used = 0;
	DrawGUI();
	frame_drawn(&fc, frames_display );
	
//	FontSetSys( F_4x6 );
}

//-------------------------------------------------------------------------
// load saved map
//-------------------------------------------------------------------------

BOOL Load( LCD_BUFFER lcd )
{
	clrscr();
  FAT_AllKeysReleased();
  GrayMode(GRAY_OFF);
  RestoreHandlerForINTs();
	
	FontSetSys( F_8x10 );
	printf( "Load what map?\n" );
	InputStr( mazename, 8 );
	if( LoadMap( mazename, lcd ) ){
		printf( "\nMaze not found" );
		ngetchx();
		LCD_restore( lcd );
		return 1;
	}
	else{
		if( (besttime = fc.frame_counter) != ULONG_MAX ){	
			printf( "\nThe best time is\n%lu", besttime/10 );
			ngetchx();
		}
	}
	fc.frame_counter = 0;
	InstallDummyHandlerForINTs();
	GrayMode( GRAY_ON );

	return 0;
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
// Start-up Functions
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------

void Splash()
{
	FAT_DisplayLogo(50);

	memset(GetPlane(LIGHT_PLANE),0,LCD_SIZE);
	memset(GetPlane(DARK_PLANE),0xff,LCD_SIZE);
	
	char map[] = { 1,1,1,1,1,1,
								 0,0,0,1,1,1,
								 1,0,1,0,1,1,
								 1,0,0,0,0,1,
								 1,1,0,1,1,1};
	int i;
	
	fc.map_width = 6;										// map_width
	fc.map_height =	5;										// map_height
	fc.map_data =	map;										// map_data
	fc.dest_plane0+=2;
	fc.dest_plane1+=2;
	fc.cam_xpos =	32;   									// cam_xpos 		(start pos_x)
	fc.cam_ypos =	96;   									// cam_ypos 		(start pos_y)
	fc.cam_orientation = ANGLE_EAST; 	// cam_orientation
	fc.frame_counter = 0;     						// frame_counter

	for( i=0; i<64/16;i++, fc.cam_xpos+=16 ) FAT_Render(&fc);
	for( i = fc.cam_orientation+ANGLE_090; fc.cam_orientation < i;
			 fc.cam_orientation+=ANGLE_015/1) FAT_Render(&fc);
	for( i=0; i<128/16; i++, fc.cam_ypos+=16 ) FAT_Render(&fc);
	for( i = fc.cam_orientation-ANGLE_090; fc.cam_orientation > i;
			 fc.cam_orientation-=ANGLE_015/1) FAT_Render(&fc);
	for( i=0; i<64/16; i++, fc.cam_xpos+=16 ) FAT_Render(&fc);
	for( i = fc.cam_orientation+ANGLE_090; fc.cam_orientation < i;
			 fc.cam_orientation+=ANGLE_015/1) FAT_Render(&fc);
	for( i=0; i<64/16; i++, fc.cam_ypos+=16 ) FAT_Render(&fc);

	fc.map_width = 16;										// map_width
	fc.map_height =	16;										// map_height
	fc.map_data =	NULL;										// map_data
	fc.dest_plane0-=2;
	fc.dest_plane1-=2;
	fc.cam_xpos =	80;   									// cam_xpos 		(start pos_x)
	fc.cam_ypos =	80;   									// cam_ypos 		(start pos_y)
	fc.cam_orientation = ANGLE_EAST+ANGLE_045; 	// cam_orientation
	fc.frame_counter = 0;     						// frame_counter
}

//-------------------------------------------------------------------------
// Prints the Title screen.
//-------------------------------------------------------------------------

void DisplayTitleScreen( int x )
{
	int i;
	char buf[20];

  for (i=0;i<2;i++) {
    SetPlane(i);
    FontSetSys( F_8x10 );
		DrawStrXY( 10, 10, "Maze Runner 3D", A_REPLACE );
		
		FontSetSys( F_6x8 );
		if( (i==0 && x!=0) || (i==1 && x==0) )
			DrawStrXY( 10, 30, "New Game", A_REPLACE );
		else
			DrawStrXY( 10, 30, "        ", A_REPLACE );
		
		if( (i==0 && x!=1) || (i==1 && x==1) )
			DrawStrXY( 10, 40, "Continue", A_REPLACE );
		else
			DrawStrXY( 10, 40, "        ", A_REPLACE );

		if( (i==0 && x!=2) || (i==1 && x==2) )
			DrawStrXY( 10, 50, "Load", A_REPLACE );
		else
			DrawStrXY( 10, 50, "    ", A_REPLACE );
			
		if( (i==0 && x!=3) || (i==1 && x==3) )
			DrawStrXY( 10, 60, "Options" , A_REPLACE );
		else
			DrawStrXY( 10, 60, "       ", A_REPLACE );

		if( (i==0 && x!=4) || (i==1 && x==4) )
			DrawStrXY( 10, 70, "Exit", A_REPLACE );
		else
			DrawStrXY( 10, 70, "    ", A_REPLACE );
  }
}

void DisplayOptionScreen( int x )
{
	int i;
	char buf[20];

  for (i=0;i<2;i++) {
    SetPlane(i);
    FontSetSys( F_8x10 );
		DrawStrXY( 10, 10, "Options", A_REPLACE );
		
		FontSetSys( F_6x8 );
		sprintf( buf, "Map Width:  %4i", fc.map_width );
		if( (i==0 && x!=0) || (i==1 && x==0) )
			DrawStrXY( 10, 30, buf , A_REPLACE );
		else
			DrawStrXY( 10, 30, "                ", A_REPLACE );

		sprintf( buf, "Map Height: %4i", fc.map_height );
		if( (i==0 && x!=1) || (i==1 && x==1) )
			DrawStrXY( 10, 40, buf , A_REPLACE );
		else
			DrawStrXY( 10, 40, "                ", A_REPLACE );

		sprintf( buf, "Night Mode: %4s", (o_night? "on": "off" ) );
		if( (i==0 && x!=2) || (i==1 && x==2) )
			DrawStrXY( 10, 50, buf , A_REPLACE );
		else
			DrawStrXY( 10, 50, "                ", A_REPLACE );

		sprintf( buf, "Comp Type: %5s", ( o_comp?(o_comp==RAND? "rand": 
																		 (o_comp==STRAT? "strgy": "fight") ): "none" ) );
		if( (i==0 && x!=3) || (i==1 && x==3) )
			DrawStrXY( 10, 60, buf, A_REPLACE );
		else
			DrawStrXY( 10, 60, "                ", A_REPLACE );

		sprintf( buf, "Timer: %9s", (o_timer? "on": "off" ) );
		if( (i==0 && x!=4) || (i==1 && x==4) )
			DrawStrXY( 10, 70, buf , A_REPLACE );
		else
			DrawStrXY( 10, 70, "                ", A_REPLACE );

		if( (i==0 && x!=5) || (i==1 && x==5) )
			DrawStrXY( 10, 80, "Done", A_REPLACE );
		else
			DrawStrXY( 10, 80, "    ", A_REPLACE );
  }
}

//-------------------------------------------------------------------------
// Displays the Title screen, sets size, gets map
//-------------------------------------------------------------------------

void Options()
{
	int x = 0;
	int d,e;
	PADSTATES key;

	Clear();	
	
	while(TRUE){
		DisplayOptionScreen( x );
	  FAT_WaitKeyPressed();
	  
		key = FAT_GetPadState();
		
		if (key & PADSTATE_ESC){
		  for(d=0;d<5000;d++) for(e=0;e<10000;e++);
		  break;
		}
		if (key & PADSTATE_UP){	
			(x>0)? (--x): (x=5);
		  for(d=0;d<5000;d++) for(e=0;e<10000;e++);
		}
		if (key & PADSTATE_DOWN){
			(x<5)? (++x): (x=0);
		  for(d=0;d<5000;d++) for(e=0;e<10000;e++);
		}
		if (key & PADSTATE_RIGHT){
			if( x==0 && fc.map_width < 250 ) fc.map_width++;
			else if( x==1 && fc.map_height < 250 ) fc.map_height++;
			else if( x==2 ) o_night = !o_night;
			else if( x==3 ){
				if( o_sprtavail )
					o_comp<3? o_comp++: (o_comp = NONE);
			}
			else if( x==4 ) o_timer = !o_timer;
		  if( x>1 )	for(d=0;d<5000;d++) for(e=0;e<10000;e++);
		  else 		  for(d=0;d<5000;d++) for(e=0;e<1000;e++);
		}
		if (key & PADSTATE_LEFT){
			if( x==0 && fc.map_width > 1 ) fc.map_width--;
			else if( x==1 && fc.map_height > 1 ) fc.map_height--;
			else if( x==2 ) o_night = !o_night;
			else if( x==3 ){
				if( o_sprtavail )
					o_comp>0? o_comp--: (o_comp = FIGHT);
			}
			else if( x==4 ) o_timer = !o_timer;
		  if( x>1 )	for(d=0;d<5000;d++) for(e=0;e<10000;e++);
		  else 		  for(d=0;d<5000;d++) for(e=0;e<1000;e++);
		}
		if (key & PADSTATE_KA){
		  for(d=0;d<5000;d++) for(e=0;e<10000;e++);
			if(  x==5 ) break;
		}
	}
	Clear();
	return;
}

//-------------------------------------------------------------------------
// Displays the Title screen, sets size, gets map
//-------------------------------------------------------------------------

BOOL Setup( LCD_BUFFER lcd )
{
	int x = 0;
	int d,e;
	PADSTATES key;

	fc.map_width	= 	8;
	fc.map_height = 	8;

	FILE *fp = fopen( "mazesave", "rb" );
	BOOL save_found = fp? TRUE: FALSE;
	fclose( fp );
	
	FontSetSys( F_8x10 );
	
	Clear();
	
	while(TRUE){
		DisplayTitleScreen( x );
	  FAT_WaitKeyPressed();
	  
		key = FAT_GetPadState();
		
		if (key & PADSTATE_ESC)   break;
		if (key & PADSTATE_UP){	
			(x>0)? (--x): (x=4);
			if( x==1 && !save_found )		--x;
		  for(d=0;d<5000;d++) for(e=0;e<10000;e++);
		}
		if (key & PADSTATE_DOWN){
			(x<4)? (++x): (x=0);
			if( x==1 && !save_found )		++x;
		  for(d=0;d<5000;d++) for(e=0;e<10000;e++);
		}
		if (key & PADSTATE_KA){
			if( x==3 )	Options();
			else				break;
		}
	}
	if( key & PADSTATE_ESC || x==4 ){
   return 1;
	}

	fc.map_width = fc.map_width*2 + 1;
	fc.map_height = fc.map_height*2 + 1;

	if( x==0 ){
		fc.map_data = (unsigned char*)calloc( fc.map_height*fc.map_width,
																				sizeof( char ) );
		if( !fc.map_data ){
			ST_helpMsg("ERROR: out of memory for map");
			return 1;
		}
		GenMaze( fc.map_data, fc.map_width, fc.map_height, 0, 1 );
	}
	else if( x==1 ){	
		if( LoadMap( "mazesave", lcd ) )	return 1;
	}
	else{
		if(Load(lcd)) 										return 1;
	}

	return 0;
}

void Congrats( int time );
void Loser();

//=============================================================================
//=============================================================================
//  main function
//=============================================================================
//=============================================================================
void _main(void)
{
	short i,j;											//loop control
	LCD_BUFFER lcd;									//lcd backup
	short oldfont = FontGetSys(); 			//used to restore previously set font

	PADSTATES  key    = 0;					//store padstate in render loop
	BOOL nomap;											//TRUE if map could not be made-ie memory
	short end;											//value of the end of the maze(half way through the last block)
	COUNTER frames_display;					//count of # of times rendered

	unsigned short* bg_img;					//buffer for background image
	unsigned short* dest;						//starts @ bg_img and is incremented for loading
	TEXCONFIG textures[2];					//wall textures
	TEXCONFIG sprite_tex[8];				//sprite textures
	FATSPRITE *sprites;							//sprites
	HANDLE sprtfile;								//file w/sprite textures
	HANDLE texfile;									//file w/wall textures
	unsigned char*  src = 0;				//pointer to textures for loading
	unsigned short* background = NULL;	//pointer to background image
	
//	fc.map_width = 16;										// map_width
//	fc.map_height =	16;										// map_height
	fc.map_data =	NULL;										// map_data
	fc.textures =	textures;								// textures
	fc.nr_sprites =	0;										// nr_sprites 	(number of sprites)
	fc.sprites = NULL;										// sprites 			(the sprite array)
//	fc.cam_xpos =	80;   									// cam_xpos 		(start pos_x)
//	fc.cam_ypos =	80;   									// cam_ypos 		(start pos_y)
//	fc.cam_orientation = ANGLE_EAST+ANGLE_045; 	// cam_orientation
	fc.dest_plane0 = NULL; 								// dest_plane0
	fc.dest_plane1 = NULL; 								// dest_plane1
	fc.background =	NULL; 								// background 	(image)
//	fc.frame_counter = 0;     						// frame_counter
	
	//-------------------------------------------------------------------------
	// initialize the FAT-Engine
	//-------------------------------------------------------------------------
	if ( (i=FAT_LoadEngine()) != FAT_LOADOKAY ) {
		switch(i) {
			case FAT_NOTGHOSTSPACE: ST_helpMsg("ERROR: client no exepacked!");break;
			case FAT_NOTFOUND:      ST_helpMsg("ERROR: no matching engine found");break;
			case FAT_LOCKFAILED:    ST_helpMsg("ERROR: locking engine failed");break;
		 	case FAT_OUTOFMEM:      ST_helpMsg("ERROR: out of memory");break;
		 	default:                ST_helpMsg("ERROR: unknown errorcode");break;
		}
		return;
	}
	
	//-------------------------------------------------------------------------
	// load textures from mazetx
	//-------------------------------------------------------------------------
	if ((texfile = FAT_GetHandleOfFile("mazetx"))) {
		if ((src = FAT_LockHandleOfFile(texfile))) {
			// load wall and door textures
			if (!FAT_LoadTextures(textures,src+2,0,2)) {
				src = 0; // just to indicate something went wrong
			}
			// load background image
			if (!src || !(background=FAT_GetGenericData(src+2,0,2304))) {
				src = 0;
			}
		}
	}

	//-------------------------------------------------------------------------
	// check if loading failed ...
	//-------------------------------------------------------------------------
	if (!texfile || !src) {
		FAT_UnloadEngine();
		if (texfile) FAT_UnlockHandleOfFile(texfile);
		ST_helpMsg("ERROR: cannot textures from mazetx");
		return;
	}
	
	//-------------------------------------------------------------------------
	// load sprite from mazesprt
	//-------------------------------------------------------------------------
	if ((sprtfile = FAT_GetHandleOfFile("mazesprt"))) {
		if ((src = FAT_LockHandleOfFile(sprtfile))) {
			// load sprite textures
			if (!FAT_LoadTextures(sprite_tex,src+2,0,8)) {
				src = 0; // just to indicate something went wrong
			}
		}																			
	}

	//-------------------------------------------------------------------------
	// check if loading failed ...
	//-------------------------------------------------------------------------
	if (!sprtfile || !src) {
		o_sprtavail = FALSE;
	}
	
	LCD_save(lcd);

	//-------------------------------------------------------------------------
	// allocate space for a background image and copy the used image into it
	//-------------------------------------------------------------------------
	if (!(bg_img = (unsigned short*)malloc(1152*2))) {
		FAT_UnloadEngine();
		FAT_UnlockHandleOfFile(texfile);
		ST_helpMsg("ERROR: out of mem");
		return;
	}
	memset(bg_img,0,1152*2);

	dest = bg_img;
	for (i=0;i<6;i++) {
		for (j=0;j<48;j++) *dest++ = 0x0000;
		for (j=0;j<48;j++) *dest++ = 0xffff;
	}

	fc.background = bg_img;

	InstallDummyHandlerForINTs();

	//-------------------------------------------------------------------------
	// initialize grayscales, clear both planes and set the destination plane
	// addresses for the FAT-Engine
	//-------------------------------------------------------------------------
	if (!GrayOn()) {
		free(bg_img);
		RestoreHandlerForINTs();
		FAT_UnloadEngine();
		FAT_UnlockHandleOfFile(texfile);
		ST_helpMsg("ERROR: out of mem");
		return;
	}

	//-------------------------------------------------------------------------
	// center verticly
	//-------------------------------------------------------------------------
	fc.dest_plane0 = GetPlane(0)+30*2;
	fc.dest_plane1 = GetPlane(1)+30*2;

	//-------------------------------------------------------------------------
	// setup calculator type specials
	//-------------------------------------------------------------------------
	if (TI89){
		display_offsetx = 0;
		display_offsety = 0;
	}
	else{
		// TI92p or V200
		display_offsetx  = 40;
		display_offsety  = 14;
		fc.dest_plane0  += (14*30+4)/2;
		fc.dest_plane1  += (14*30+4)/2;
	}

	Splash();
	nomap = Setup(lcd);

	if( o_night ){
		dest = bg_img;
		for(i=0;i<6;i++){
			unsigned short* src = background+i;
			for(j=0;j<96;j++,dest++,src+=6){
				*dest = *src;
				*(dest+576) = *(src+576);
			}
		}
		//-------------------------------------------------------------------------
		// set predraw callback function which is called before the vertical strips
		// get rendered, but after the DEPTHDATA structure is filled ...
		//-------------------------------------------------------------------------
		FAT_SetPredrawCallback(PredrawCallback);
//		fc.background = bg_img;
	}

	if( o_comp ){
		short sX = o_comp==FIGHT? (fc.map_width+1)/2/2: 1, 
					sY = o_comp==FIGHT? (fc.map_height+1)/2/2: 1;
					
		if( !loaded ){
			fc.nr_sprites = o_comp==FIGHT? sX*sY-1: 1;
			fc.sprites = calloc( fc.nr_sprites, sizeof(FATSPRITE) );
		}
		sprites = fc.sprites;

		for( i=0; i<fc.nr_sprites; i++ ){
			if(!loaded)	FAT_InitSprite(sprites+i);
			sprites[i].drawmode			= SPRITEMODE_SIZE64;
			sprites[i].nr_textures	= 8;
			sprites[i].textures			= sprite_tex;
		}
		if( !loaded ){
			int n=0;
			for( i=0; i<sX; i++ ){
				for( j=0; j<sY; j++ ){ 
					if( o_comp!=FIGHT || i!=0 || j!=0 ){
						sprites[n].orientation	= ANGLE_SOUTH;
						sprites[n].xpos					= (i*2*2+1)*64+32;
						sprites[n].ypos					= (j*2*2+1)*64+32;
						n++;
					}
				}
			}
		}
		
		randomize();
	}

	if( !nomap ){
		Clear();
		InitGUI( &frames_display );
		
		//-------------------------------------------------------------------------
		// the main rendering loop ...
		//-------------------------------------------------------------------------
		end = 64*fc.map_width - 32;
		do{
			FAT_Render(&fc);   								// render image 
			frame_drawn(&fc,&frames_display);	// handles frame counter drawing
			if( o_comp ){
				MoveSprite();
				if( o_comp != FIGHT && fc.sprites->xpos+1 >= end ){
					fc.sprites->xpos++;
					break;
				}
			}

			if( (key = FAT_GetPadState()) ) {
				if( key & PADSTATE_ESC )   break;
				
				if( key & PADSTATE_UPDOWN )	PlayerMove(&fc, key & PADSTATE_UP, 0);
				if( key & PADSTATE_LEFTRIGHT ){
					// STRAFE KEY: <> (TI-89) / F2 (TI-92p)
					if( key & PADSTATE_KB )
						PlayerMove(&fc,key&PADSTATE_RIGHT,1);
					else
						PlayerTurn(&fc,key&PADSTATE_RIGHT,ANGLE_015/2);
				}
				if( key & PADSTATE_APPS )
					DisplayMiniMap(fc.cam_xpos, fc.cam_ypos, &frames_display );
				if( key & PADSTATE_KD )
					Save(&frames_display);
				if( o_comp & FIGHT && key & PADSTATE_KA )
					Shoot();
			}
		}while( fc.cam_xpos < end );

		Clear();
		if( fc.cam_xpos >= end )										// if you got to the end
			Congrats( fc.frame_counter );						// display rank
		else if(o_comp? fc.sprites->xpos >= end: FALSE )
			Loser();
		else
			SaveMap( "mazesave" );									// else save progress
	}
	//-------------------------------------------------------------------------
	// some clean up stuff ...
	//-------------------------------------------------------------------------

	FAT_AllKeysReleased();
	GrayMode(GRAY_OFF);
	RestoreHandlerForINTs();
	free( bg_img );
	if( o_comp )			free(sprites);
	if( !nomap )			free(fc.map_data);

	LCD_restore(lcd);
	ST_helpMsg(FAT_GetPoweredByString());

	FAT_UnloadEngine();
	FAT_UnlockHandleOfFile( texfile );
	if( o_sprtavail ) FAT_UnlockHandleOfFile( sprtfile );
	OSqclear(kbd_queue());
	FontSetSys(oldfont);
}

//-------------------------------------------------------------------------
// Congratulations
//-------------------------------------------------------------------------

void Congrats( int time )
{
	FILE* fp = fopen( mazename, "r+b" );
	short buf[5];
    
	FontSetSys(F_8x10);
	clrscr();    

	printf_xy( 5, 40, "Rank:" );

	if( time < fc.map_width*fc.map_height )
		printf_xy( 5, 55, "Lighting" );
	else if( time < fc.map_width*fc.map_height * 2 )
		printf_xy( 5, 55, "Sprinter" );
	else if( time < fc.map_width*fc.map_height * 3 )
		printf_xy( 5, 55, "Power Walker" );
	else
		printf_xy( 5, 55, "Couch Potato" );
		
	if( besttime ){
		if( fc.frame_counter < besttime ){
			printf( "You set a new best time of:\n%lu", fc.frame_counter/10 );
			fread( buf, 5, sizeof(short), fp );
			fwrite( &fc.frame_counter, 1, sizeof(long), fp );
		}
		else 
			printf( "Too slow" );
	}
	FAT_AllKeysReleased();
	FAT_WaitKeyPressed();
}

void Loser()
{
	FontSetSys(F_8x10);
	clrscr();    

	printf_xy( 5, 20, "You were beaten by" );
	printf_xy( 5, 35, "  %8s Sim.", o_comp == RAND? "Random": "Strategy" );
	printf_xy( 5, 50, "  Computers Are");
	printf_xy( 5, 65, "    Superior!" );

	FAT_AllKeysReleased();
	FAT_WaitKeyPressed();
}

//-------------------------------------------------------------------------
// History
//-------------------------------------------------------------------------

// Started on 3/16/2002; 9:24:58 PM

// version 0.1 beta			3/23/2002
//   first version
//	 random 16x16 maze

// version 0.2 beta			3/30/2002
//   congratulations
//   time
//	 print map
// 	 title screen
//	 select size
//   added the too close to wall fix from demo 3

// version 0.3 beta			4/9/2002
//	First public release!
//   reorganized code
//	 rewrote to have fc as only global
//	 rewrote congrats: rank
// 	 external files:
//  	save game

// version 0.4 beta			5/4/2002
//  Second release.
//   now uses FAT v. 1.00
//   switched from -Os tor -O3 for better speed
//	 night mode
//	 race against a random or stratagy sprite
//	 ability to shut timer off
//	 added options screen
//   more complete saved game files- night, sprite, timer

// version 0.5 beta
//	not keeping this up-to-date any more, see read me

//-------------------------------------------------------------------------
// Future
//-------------------------------------------------------------------------

// chase blocks

// fighting

// better looking interface

// external files:
// 	options

// external maps - external editor
